'use strict'

// mode ref => https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/globalCompositeOperation
modules['overlay'] = {
    config: [
        ['alpha', 'float', 1],
        ['position', 'float2', [0, 0]],
        ['source', 'string', ''],
        ['mode', 'string', 'source-over'],
        ['scale', 'float', 1.0],
        ['width', 'int', 640],
        ['height', 'int', 480],
    ],

    render: async (time, self, config) => {
        const {
            alpha,
            position,
            source,
            mode,
            scale,
            width,
            height,
        } = { ...config }

        if (source && alpha > 0) {
            const destWidth = Math.floor(width * scale)
            const destHeight = Math.floor(height * scale)
            const posX = position[0] - Math.floor((destWidth - width) / 2)
            const posY = position[1] - Math.floor((destHeight - height) / 2)
    
            const src = getColourTexture(source)
            if (scale !== 1) {
                const iScale = 1 / scale
                for (let y = 0; y < 480; ++y) {
                    const srcY = (((y-240)*iScale)+240 - position[1]) >> 0
                    if (srcY >= 0 && srcY < 480 )
                        for (let x = 0; x < 640; ++x) {
                        const srcX = (((x-320)*iScale)+320 - position[0]) >> 0
                        if (srcX >= 0 && srcX < 640 ) {
                            let srcIndex = (srcX + srcY*640) * 4
                            let dstIndex = (x + y*640) * 4
                            const srcR = src.data[srcIndex++]
                            const srcG = src.data[srcIndex++]
                            const srcB = src.data[srcIndex++]
                            const srcA = src.data[srcIndex++]
                            const a = alpha * srcA/255
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcR, a); dstIndex++
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcG, a); dstIndex++
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcB, a); dstIndex++
                        }
                    }
                }
            } else if (position[0] !== 0 || position[1] !== 0) {
                for (let y = 0; y < 480; ++y) {
                    const srcY = y - position[1] >> 0
                    if (srcY >= 0 && srcY < 480 )
                        for (let x = 0; x < 640; ++x) {
                        const srcX = x - position[0] >> 0
                        if (srcX >= 0 && srcX < 640 ) {
                            let srcIndex = (srcX + srcY*640) * 4
                            let dstIndex = (x + y*640) * 4
                            const srcR = src.data[srcIndex++]
                            const srcG = src.data[srcIndex++]
                            const srcB = src.data[srcIndex++]
                            const srcA = src.data[srcIndex++]
                            const a = alpha * srcA/255
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcR, a); dstIndex++
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcG, a); dstIndex++
                            workingBuffer.data[dstIndex] = lerp(workingBuffer.data[dstIndex], srcB, a); dstIndex++
                        }
                    }
                }
            }
            else if (alpha <= 0) {
            } else if (alpha === 1) {
                let s = 0
                let d = 0
                for (let i = 0; i < 640 * 480; ++i) {
                    const srcR = src.data[s++]
                    const srcG = src.data[s++]
                    const srcB = src.data[s++]
                    const srcA = src.data[s++]
                    const a = srcA/255
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcR, a); d++
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcG, a); d++
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcB, a); d++
                    d++
                }
            } else {
                let s = 0
                let d = 0
                for (let i = 0; i < 640 * 480; ++i) {
                    const srcR = src.data[s++]
                    const srcG = src.data[s++]
                    const srcB = src.data[s++]
                    const srcA = src.data[s++]
                    const a = alpha * srcA/255
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcR, a); d++
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcG, a); d++
                    workingBuffer.data[d] = lerp(workingBuffer.data[d], srcB, a); d++
                    d++
                }
            }
        }
    }
}